require(serialize)
require(gdbm)
envlist <- function(e) {
    names <- ls(e, all=TRUE)
    list <- lapply(names, get, env=e, inherits=FALSE)
    names(list) <- names
    list
}
listIntoEnv <- function(list, e) {
    names <- names(list)
    for (i in seq(along = names))
        assign(names[i], list[[i]], env = e)
}
envtable <- function() {
    idx <- 0
    envs <- NULL
    enames <- character(0)
    find <- function(v, keys, vals)
        for (i in seq(along=keys))
            if (identical(v, keys[[i]]))
                return(vals[i])
    getname <- function(e) find(e, envs, enames)
    getenv <- function(n) find(n, enames, envs)
    insert <- function(e) {
        idx <<- idx + 1
        name <- paste("env", idx, sep="::")
        envs <<- c(e, envs)
        enames <<- c(name, enames)
        name
    }
    list(insert = insert, getenv = getenv, getname = getname)
}
makeGdbmWriter <- function(file) {
    force(file)
    gdbmNew(file)
    list(insert = function(name, value) gdbmInsert(file, name, value))
}
makeGdbmReader <- function(file) {
    force(file)
    list(list = function() gdbmList(file),
         fetch = function(name) gdbmFetch(file, name),
         exists = function(name) gdbmExists(file, name))
}
makeShelfWriter <- function(db,  ascii = TRUE) {
    if (is.character(db))
        db <- makeGdbmWriter(db)
    table <- envtable()
    ser <- function(val)
        serialize(val, NULL, ascii = ascii, refhook = envhook)
    envhook <- function(e) {
        if (is.environment(e)) {
            name <- table$getname(e)
            if (is.null(name)) {
                name <- table$insert(e)
                data <- list(bindings = envlist(e),
                             enclos = parent.env(e))
                db$insert(name, ser(data))
            }
            name
        }
    }
    putvar <- function(name, val, prefix="var", sep="::") {
        key <- paste("var", name, sep="::")
        db$insert(key, ser(val))
    }
    list(putvar = putvar)
}

makeShelf <- function(list, db, ascii = TRUE)
{
    names <- names(list)
    if (length(names) != length(list))
        stop("must provide a named list of values")
    s <- makeShelfWriter(db, ascii)
    for (i in seq(along=names))
        s$putvar(names[i], list[[i]])
}
listShelf <- function(db) {
    if (is.character(db))
        db <- makeGdbmReader(db)
    prefpat <- "^var::"
    sub(prefpat, "", grep(prefpat,db$list(), value=TRUE))
}
openShelf <- function(db) {
    if (is.character(db))
        db <- makeGdbmReader(db)
    envenv <- new.env(hash = TRUE)
    varkey <- function(name) paste("var", name, sep="::")
    fetch <- function(name)
        unserialize(db$fetch(name), refhook = envhook)
    envhook <- function(n) {
       if (exists(n, env = envenv, inherits = FALSE))
           get(n, env = envenv, inherits = FALSE)
       else {
           e <- new.env(hash = TRUE)
           assign(n, e, env = envenv) # MUST do this immediately
           data <- fetch(n)
           parent.env(e) <- data$enclos
           listIntoEnv(data$bindings, e)
           e
        }
    }
    list(getvar = function(name) fetch(varkey(name)),
         exists = function(name) db$exists(varkey(name)),
         list = function() listShelf(db))
}
loadShelf <- function(db, envir = parent.frame()) {
    s <- openShelf(db)
    wrap<-function(name) {
        name <- name # need to force evaluation!
        delay(s$getvar(name), env=environment())
    }
    for (n in listShelf(db))
        assign(n, wrap(n), envir = envir)
}
attachShelf <- function(db, pos = 2, name) {
    if (missing(name)) {
        if (is.character(db))
            name <- paste("shelf", db, sep = ":")
        else name = "shelf:<no name>"
    }
    env <- attach(NULL, pos = pos, name = name)
    on.exit(detach(pos))
    loadShelf(db, env)
    on.exit()
}
